/*  
    Copyright (C) 2007 Rikiya Ayukawa <ayukawa@sde.cs.titech.ac.jp>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
 
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
*/

#define _GNU_SOURCE

#include <ruby.h>
#include <node.h>

#include <dlfcn.h>
#include <string.h>
#include <signal.h>

#define PATH_SEPARATOR '/'
#define HOOK_EVENT_NAME "break"

extern int bdg_tracing;
extern VALUE bdg_trace_func;

///ruby上のArray. rubyスレッドのcritical値を記録する変数
static VALUE bdsDbgCriticalStack;

VALUE bdrb_is_extension(VALUE self, VALUE klass, VALUE id);
VALUE bdrb_rdb2gdb(VALUE self);

static int bd_start_switch(){
  VALUE crit = rb_funcall(rb_cThread, rb_intern("critical"), 0);
  rb_funcall(bdsDbgCriticalStack, rb_intern("push"), 1, crit);
  rb_funcall(rb_cThread, rb_intern("critical="), 1, Qtrue);

  return 1;
}

static int bd_end_switch(){
  VALUE crit = rb_funcall(bdsDbgCriticalStack, rb_intern("pop"), 0);
  rb_funcall(rb_cThread, rb_intern("critical="), 1, crit);

  return 1;
}

int bdcmd_switch(){
  VALUE event_name = rb_str_new2(HOOK_EVENT_NAME);
  bdg_tracing++;
  bd_start_switch();
  rb_funcall(bdg_trace_func, rb_intern("call"), 6, 
	     event_name, Qnil, Qnil, Qnil, Qnil, Qnil);
  bd_end_switch();
  bdg_tracing--;

  return 1;
}

// rdbからgdbへのstep実行遷移のための関数
static void bdbrk_r2gstep(void* breakpoint_addr){
}

static const char* bd_get_filename_from_addr(void* addr){
  Dl_info info;

  if(dladdr(addr, &info) == 0){
    return NULL; //addrなんてアドレスしらん
  }

  return info.dli_fname;
}


/**
 * C拡張ライブラリをmapした仮想メモリ空間上のページに存在する関数か？
 * @param addr 関数アドレス
 * @return 判定がtrueなら1, そうでなければ0
 */
static int bd_is_extension(void* addr){
  const char* libfile = bd_get_filename_from_addr(addr);

  //無効なメモリ領域 or heapなどの実行命令コード領域外
  if(libfile == NULL) return 0;

  //共有オブジェクトがlibrubyだったら、ruby自体の共有オブジェクトなので
  //拡張ライブラリではない
  if(strstr(basename(libfile), "libruby") != NULL) return 0;
  
  return 1;
}


/**
 * メソッドとして呼び出すC関数がC拡張ライブラリであるか判定するC拡張関数
 * @return C拡張ライブラリならQtrue, そうでなければQfalse
 * @sa bdbrk_r2gstep
 */
VALUE bdrb_is_extension(VALUE self, VALUE klass, VALUE id){
  ID ide;
  VALUE k;
  NODE* body;
  Check_Type(id, T_SYMBOL);
  if(!SYMBOL_P(id)) return Qfalse;
  k = CLASS_OF(klass);
  ide = SYM2ID(id);
  body = rb_method_node(k, ide);

  if(body == NULL) return Qfalse;
  switch(nd_type(body)){
  case NODE_CFUNC:
    if(bd_is_extension(body->nd_cfnc)){
      bdbrk_r2gstep(body->nd_cfnc);
      return Qtrue;
    }
    //    return INT2FIX(body->nd_cfnc);
    break;
  default:
    break;
  }

  return Qfalse;
}

VALUE bdrb_rdb2gdb(VALUE self){
  raise(SIGTRAP);
  return Qnil;
}


/**
 * ruby/eval.cのtracing, trace_func変数はstaticなので、
 * ruby.soのシンボルテーブルに存在しないことがある
 * 自前でruby-event-hookを行い、自前のtracing, trace_func変数を
 * global変数として用意することで、シンボルテーブルにのっける
 */

/***************/
/*  global変数 */
/***************/
int bdg_tracing = 0;
VALUE bdg_trace_func = Qnil;

/***************/
/*  関数宣言   */
/***************/
void call_trace_func(rb_event_t event, NODE* node, VALUE self, ID id, VALUE klass);

static const char *
bd_get_event_name(rb_event_t event);

static VALUE bdrb_set_trace_func(VALUE obj, VALUE trace)
{

    rb_secure(4);
    if (NIL_P(trace)) {
	bdg_trace_func = 0;
	rb_remove_event_hook(call_trace_func);
	return Qnil;
    }

    /*rb_obj_is_procはeval.cのstatic関数（かつ、この関数内部でstatic変数を使っている)
    if (!rb_obj_is_proc(trace)) {
	rb_raise(rb_eTypeError, "trace_func needs to be Proc");
    }
    */
    bdg_trace_func = trace;
    /*event_hooksはeval.c内のstatic変数
    for (hook = event_hooks; hook; hook = hook->next) {
	if (hook->func == call_trace_func)
	    return trace;
    }
    */
    rb_add_event_hook(call_trace_func, RUBY_EVENT_ALL);
    return trace;
}

/**
 * 基本的call_trace_func@ruby/eval.cのパクリ
 * gdbのcallコマンドでこの関数を呼べる用にするため、わざとstatic修飾子を
 * つけてない
 */
void call_trace_func(rb_event_t event, NODE* node, VALUE self, ID id, VALUE klass)
{
  VALUE srcfile;
  const char* event_name;
  int my_sourceline;
  const char* my_sourcefile = NULL;
  if(!bdg_trace_func) return;
  if(bdg_tracing) return;
  //  if (ruby_in_compile) return;
  if(id == ID_ALLOCATOR) return;

  //if (!(node_save = ruby_current_node)) {
  //  node_save = NEW_NEWLINE(0);
  //}

  bdg_tracing = 1;
  //prev = ruby_frame;
  //PUSH_FRAME();
  //*ruby_frame = *prev;
  //ruby_frame->prev = prev;
  //ruby_frame->iter = 0;	/* blocks not available anyway */
  if(node){
    //	ruby_current_node = node;
    //	ruby_frame->node = node;
    my_sourcefile = node->nd_file;
    my_sourceline = nd_line(node);
  }
  if(klass){
    if(TYPE(klass) == T_ICLASS){
      klass = RBASIC(klass)->klass;
    }
    else if (FL_TEST(klass, FL_SINGLETON)){
      klass = rb_iv_get(klass, "__attached__");
    }
  }
  //PUSH_TAG(PROT_NONE);
  //  raised = thread_reset_raised();
  //  if ((state = EXEC_TAG()) == 0) {
	srcfile = rb_str_new2(my_sourcefile?my_sourcefile:"(ruby)");
	event_name = bd_get_event_name(event);

	rb_funcall(bdg_trace_func, rb_intern("call"), 6,
		   rb_str_new2(event_name),
		   srcfile,
		   INT2FIX(my_sourceline),
		   id?ID2SYM(id):Qnil,
		   self?rb_funcall2(self, rb_intern("binding"), 0, NULL):Qnil,
		   klass);
      
	//}
    //if (raised) thread_set_raised();
    //POP_TAG();
    //POP_FRAME();

    bdg_tracing = 0;
    //ruby_current_node = node_save;
    //SET_CURRENT_SOURCE();
    //    if (state) JUMP_TAG(state);

}

static const char *
bd_get_event_name(rb_event_t event)
{
    switch (event) {
    case RUBY_EVENT_LINE:
	return "line";
    case RUBY_EVENT_CLASS:
	return "class";
    case RUBY_EVENT_END:
	return "end";
    case RUBY_EVENT_CALL:
	return "call";
    case RUBY_EVENT_RETURN:
	return "return";
    case RUBY_EVENT_C_CALL:
	return "c-call";
    case RUBY_EVENT_C_RETURN:
	return "c-return";
    case RUBY_EVENT_RAISE:
	return "raise";
    default:
	return "unknown";
    }
}

void Init_bdrb(void){
  VALUE mBDRB;
  mBDRB = rb_define_module("BDRB");
  
  rb_define_module_function(mBDRB, "is_extension", &bdrb_is_extension, 2);
  rb_define_module_function(mBDRB, "rdb2gdb", &bdrb_rdb2gdb, 0);
  rb_define_module_function(mBDRB, "set_trace_func", &bdrb_set_trace_func, 1);

  bdsDbgCriticalStack = rb_ary_new();
  rb_global_variable(&bdsDbgCriticalStack);
}
